@@ -63,6 +63,11 @@ class UASocketClient(object):
             if callback:
                 future.add_done_callback(callback)
             self._callbackmap[self._request_id] = future
+
+            # Change to the new security token if the connection has been renewed.
+            if self._connection.next_security_token.TokenId != 0:
+                self._connection.revolve_tokens()
+
             msg = self._connection.message_to_binary(binreq, message_type=message_type, request_id=self._request_id)
             self._socket.write(msg)
         return future
@@ -96,9 +101,11 @@ class UASocketClient(object):
                 self._receive()
             except ua.utils.SocketClosedException:
                 self.logger.info("Socket has closed connection")
+                self._connection.close()
                 break
             except UaError:
                 self.logger.exception("Protocol Error")
+        self._cancel_all_callbacks()
         self.logger.info("Thread ended")
 
     def _receive(self):
@@ -125,6 +132,12 @@ class UASocketClient(object):
                 )
         future.set_result(body)
 
+    def _cancel_all_callbacks(self):
+        for request_id, fut in self._callbackmap.items():
+            self.logger.info("Cancelling request {:d}".format(request_id))
+            fut.cancel()
+        self._callbackmap.clear()
+
     def _create_request_header(self, timeout=1000):
         hdr = ua.RequestHeader()
         hdr.AuthenticationToken = self.authentication_token
@@ -161,7 +174,7 @@ class UASocketClient(object):
         self.logger.info("Socket closed, waiting for receiver thread to terminate...")
         if self._thread and self._thread.is_alive():
             self._thread.join()
-        self._callbackmap.clear()
+        self._cancel_all_callbacks()
         self.logger.info("Done closing socket: Receiving thread terminated, socket disconnected")
 
     def send_hello(self, url, max_messagesize=0, max_chunkcount=0):
@@ -183,11 +196,9 @@ class UASocketClient(object):
         request.Parameters = params
         future = self._send_request(request, message_type=ua.MessageType.SecureOpen)
 
-        # FIXME: we have a race condition here
-        # we can get a packet with the new token id before we reach to store it..
         response = struct_from_binary(ua.OpenSecureChannelResponse, future.result(self.timeout))
         response.ResponseHeader.ServiceResult.check()
-        self._connection.set_channel(response.Parameters)
+        self._connection.set_channel(response.Parameters, params.RequestType, params.ClientNonce)
         return response.Parameters
 
     def close_secure_channel(self):
@@ -621,7 +632,7 @@ class UaClient(object):
         response.ResponseHeader.ServiceResult.check()
         # nothing to return for this service
 
-    def get_attribute(self, nodes, attr):
+    def get_attributes(self, nodes, attr):
         self.logger.info("get_attribute")
         request = ua.ReadRequest()
         for node in nodes:
@@ -633,3 +644,22 @@ class UaClient(object):
         response = struct_from_binary(ua.ReadResponse, data)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
+
+    def set_attributes(self, nodeids, datavalues, attributeid=ua.AttributeIds.Value):
+        """
+        Set an attribute of multiple nodes
+        datavalue is a ua.DataValue object
+        """
+        self.logger.info("set_attributes of several nodes")
+        request = ua.WriteRequest()
+        for idx, nodeid in enumerate(nodeids):
+            attr = ua.WriteValue()
+            attr.NodeId = nodeid
+            attr.AttributeId = attributeid
+            attr.Value = datavalues[idx]
+            request.Parameters.NodesToWrite.append(attr)
+        data = self._uasocket.send_request(request)
+        response = struct_from_binary(ua.WriteResponse, data)
+        response.ResponseHeader.ServiceResult.check()
+        return response.Results
+
